@@ -1,12 +1,9 @@
 import errno
-import glob
 import os
 import re
-import shutil
 import tempfile
 import typing
 
-import audfactory
 import audeer
 
 from audbackend.core import utils
@@ -46,12 +43,15 @@ class Backend:
             self,
             path: str,
             version: str,
+            *,
+            ext: str = None,
     ) -> str:
         r"""Get MD5 checksum for file on backend.
 
         Args:
             path: path to file on backend
             version: version string
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             MD5 checksum
@@ -60,7 +60,7 @@ class Backend:
             FileNotFoundError: if file does not exist on backend
 
         """
-        backend_path = self.path(path, version)
+        backend_path = self.path(path, version, ext=ext)
 
         if not self._exists(backend_path):
             raise FileNotFoundError(
@@ -80,18 +80,21 @@ class Backend:
             self,
             path: str,
             version: str,
+            *,
+            ext: str = None,
     ) -> bool:
         r"""Check if file exists on backend.
 
         Args:
             path: path to file on backend
             version: version string
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             ``True`` if file exists
 
         """
-        path = self.path(path, version)
+        path = self.path(path, version, ext=ext)
         return self._exists(path)
 
     def get_archive(
@@ -138,6 +141,8 @@ class Backend:
             src_path: str,
             dst_path: str,
             version: str,
+            *,
+            ext: str = None,
     ):
         r"""Get file from backend.
 
@@ -145,6 +150,7 @@ class Backend:
             src_path: path to file on backend
             dst_path: destination path to local file
             version: version string
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             full path to local file
@@ -153,7 +159,7 @@ class Backend:
             FileNotFoundError: if file does not exist on backend
 
         """
-        src_path = self.path(src_path, version)
+        src_path = self.path(src_path, version, ext=ext)
         if not self._exists(src_path):
             raise FileNotFoundError(
                 errno.ENOENT, os.strerror(errno.ENOENT), src_path,
@@ -167,6 +173,7 @@ class Backend:
     def _glob(
             self,
             pattern: str,
+            folder: typing.Optional[str],
     ) -> typing.List[str]:  # pragma: no cover
         r"""Return matching files names."""
         raise NotImplementedError()
@@ -174,6 +181,8 @@ class Backend:
     def glob(
             self,
             pattern: str,
+            *,
+            folder: str = None,
     ) -> typing.List[str]:
         r"""Return matching files names.
 
@@ -181,12 +190,13 @@ class Backend:
 
         Args:
             pattern: pattern string
+            folder: search under this folder
 
         Returns:
             matching files on backend
 
         """
-        return self._glob(pattern)
+        return self._glob(pattern, folder)
 
     def join(
             self,
@@ -203,16 +213,22 @@ class Backend:
             path joined by :attr:`Backend.sep`
 
         """
-        return self.sep.join([path] + [p for p in paths])
+        paths = [path] + [p for p in paths]
+        # skip part if '' or None
+        paths = [path for path in paths if path]
+        return self.sep.join(paths)
 
     def latest_version(
             self,
             path: str,
+            *,
+            ext: str = None,
     ) -> str:
         r"""Latest version of a file.
 
         Args:
             path: relative path to file in repository
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             version string
@@ -221,7 +237,7 @@ class Backend:
             RuntimeError: if file does not exist on backend
 
         """
-        vs = self.versions(path)
+        vs = self.versions(path, ext=ext)
         if not vs:
             raise RuntimeError(
                 f"Cannot find a version for "
@@ -244,6 +260,8 @@ class Backend:
             self,
             path: str,
             version: str,
+            *,
+            ext: str = None,
     ) -> str:
         r"""File path on backend.
 
@@ -254,15 +272,28 @@ class Backend:
         Args:
             path: relative path to file in repository
             version: version string
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             file path on backend
 
+        Raises:
+            ValueError: if ``path`` contains invalid character
+            ValueError: if ``path`` does not end on file extension
+
         Example:
-            >>> backend = FileSystem('~/my-host', 'data')
+            >>> import audbackend
+            >>> backend = audbackend.FileSystem('~/my-host', 'data')
             >>> path = backend.path('media/archive1.zip', '1.0.0')
             >>> os.path.basename(path)
             'archive1-1.0.0.zip'
+            >>> path = backend.path(
+            ...     'media/archive1.tar.gz',
+            ...     '1.0.0',
+            ...     ext='tar.gz',
+            ... )
+            >>> os.path.basename(path)
+            'archive1-1.0.0.tar.gz'
 
         """
         allowed_chars = re.compile(BACKEND_ALLOWED_CHARS)
@@ -272,7 +303,19 @@ class Backend:
                 f"allowed characters are '{BACKEND_ALLOWED_CHARS}'."
             )
         folder, file = self.split(path)
-        name, ext = os.path.splitext(file)
+        if ext is None:
+            name, ext = os.path.splitext(file)
+        elif ext == '':
+            name = file
+        else:
+            if not ext.startswith('.'):
+                ext = '.' + ext
+            if not path.endswith(ext):
+                raise ValueError(
+                    f"Invalid path name '{path}', "
+                    f"does not end on '{ext}'."
+                )
+            name = file[:-len(ext)]
         return self._path(folder, name, ext, version)
 
     def put_archive(
@@ -336,6 +379,8 @@ class Backend:
             src_path: str,
             dst_path: str,
             version: str,
+            *,
+            ext: str = None,
     ):
         r"""Put file on backend.
 
@@ -347,6 +392,7 @@ class Backend:
             src_path: path to local file
             dst_path: path to file on backend
             version: version string
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             file path on backend
@@ -360,7 +406,7 @@ class Backend:
                 errno.ENOENT, os.strerror(errno.ENOENT), src_path,
             )
 
-        dst_path = self.path(dst_path, version)
+        dst_path = self.path(dst_path, version, ext=ext)
 
         # skip if file with same checksum exists on backend
         skip = self._exists(dst_path) and \
@@ -381,12 +427,15 @@ class Backend:
             self,
             path: str,
             version: str,
+            *,
+            ext: str = None,
     ) -> str:
         r"""Remove file from backend.
 
         Args:
             path: path to file on backend
             version: version string
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             path of removed file on backend
@@ -395,7 +444,7 @@ class Backend:
             FileNotFoundError: if file does not exist on backend
 
         """
-        path = self.path(path, version)
+        path = self.path(path, version, ext=ext)
         if not self._exists(path):
             raise FileNotFoundError(
                 errno.ENOENT, os.strerror(errno.ENOENT), path,
@@ -438,290 +487,20 @@ class Backend:
     def versions(
             self,
             path: str,
+            *,
+            ext: str = None,
     ) -> typing.List[str]:
         r"""Versions of a file.
 
         Args:
             path: path to file on backend
+            ext: file extension, if ``None`` uses characters after last dot
 
         Returns:
             list of versions in ascending order
 
         """
         folder, file = self.split(path)
-        name = audeer.basename_wo_ext(file)
+        name = audeer.basename_wo_ext(file, ext=ext)
         vs = self._versions(folder, name)
         return audeer.sort_versions(vs)
-
-
-class Artifactory(Backend):
-    r"""Artifactory backend.
-
-    Stores files and archives on Artifactory.
-
-    Args:
-        host: host address
-        repository: repository name
-
-    """
-
-    def __init__(
-            self,
-            host,
-            repository,
-    ):
-        super().__init__(host, repository)
-
-    def _checksum(
-            self,
-            path: str,
-    ) -> str:
-        r"""MD5 checksum of file on backend."""
-        return audfactory.checksum(path)
-
-    def _path(
-            self,
-            folder: str,
-            name: str,
-            ext: str,
-            version: str,
-    ) -> str:
-        r"""File path on backend."""
-        server_url = audfactory.url(
-            self.host,
-            repository=self.repository,
-            group_id=audfactory.path_to_group_id(folder),
-            name=name,
-            version=version,
-        )
-        return f'{server_url}/{name}-{version}{ext}'
-
-    def _exists(
-            self,
-            path: str,
-    ) -> bool:
-        r"""Check if file exists on backend."""
-        try:
-            # Can lead to RuntimeError: 404 page not found
-            return audfactory.path(path).exists()
-        except RuntimeError:  # pragma: nocover
-            return False
-
-    def _get_file(
-            self,
-            src_path: str,
-            dst_path: str,
-    ):
-        r"""Get file from backend."""
-        audfactory.download(src_path, dst_path, verbose=False)
-
-    def _glob(
-            self,
-            pattern: str,
-    ) -> typing.List[str]:
-        r"""Return matching files names."""
-        url = audfactory.url(
-            self.host,
-            repository=self.repository,
-        )
-        path = audfactory.path(url)
-        try:
-            result = [str(x) for x in path.glob(pattern)]
-        except RuntimeError:  # pragma: no cover
-            result = []
-        return result
-
-    def _put_file(
-            self,
-            src_path: str,
-            dst_path: str,
-    ):
-        r"""Put file to backend."""
-        audfactory.deploy(src_path, dst_path)
-
-    def _remove_file(
-            self,
-            path: str,
-    ):
-        r"""Remove file from backend."""
-        audfactory.path(path).unlink()
-
-    def _versions(
-            self,
-            folder: str,
-            name: str,
-    ) -> typing.List[str]:
-        r"""Versions of a file."""
-        group_id = audfactory.path_to_group_id(folder)
-        return audfactory.versions(self.host, self.repository, group_id, name)
-
-
-class FileSystem(Backend):
-    r"""File system backend.
-
-    Stores files and archives on a file system.
-
-    Args:
-        host: host directory
-        repository: repository name
-
-    """
-    def __init__(
-            self,
-            host: str,
-            repository: str,
-    ):
-        super().__init__(audeer.safe_path(host), repository)
-
-    def _checksum(
-            self,
-            path: str,
-    ) -> str:
-        r"""MD5 checksum of file on backend."""
-        return utils.md5(path)
-
-    def _path(
-            self,
-            folder: str,
-            name: str,
-            ext: str,
-            version: str,
-    ) -> str:
-        r"""File path on backend."""
-        path = os.path.join(
-            self.host,
-            self.repository,
-            folder.replace(self.sep, os.path.sep),
-            name,
-        )
-        if version is not None:
-            path = os.path.join(
-                path,
-                version,
-                f'{name}-{version}{ext}',
-            )
-        return path
-
-    def _exists(
-            self,
-            path: str,
-    ) -> bool:
-        r"""Check if file exists on backend."""
-        return os.path.exists(path)
-
-    def _get_file(
-            self,
-            src_path: str,
-            dst_path: str,
-    ):
-        r"""Get file from backend."""
-        shutil.copy(src_path, dst_path)
-
-    def _glob(
-            self,
-            pattern: str,
-    ) -> typing.List[str]:
-        r"""Return matching files names."""
-        pattern = pattern.replace(self.sep, os.path.sep)
-        root = os.path.join(self.host, self.repository)
-        path = os.path.join(root, pattern)
-        return [os.path.join(root, p) for p in glob.glob(path, recursive=True)]
-
-    def _put_file(
-            self,
-            src_path: str,
-            dst_path: str,
-    ):
-        r"""Put file to backend."""
-        audeer.mkdir(os.path.dirname(dst_path))
-        shutil.copy(src_path, dst_path)
-
-    def _remove_file(
-            self,
-            path: str,
-    ):
-        r"""Remove file from backend."""
-        os.remove(path)
-
-    def _versions(
-            self,
-            folder: str,
-            name: str,
-    ) -> typing.List[str]:
-        r"""Versions of a file."""
-        root = self._path(folder, name, '', None)
-        vs = []
-        if os.path.exists(root):
-            vs = [
-                v for v in os.listdir(root)
-                if os.path.isdir(os.path.join(root, v))
-            ]
-        return vs
-
-
-backends = {}
-r"""Backend cache."""
-
-backend_registry = {
-    'file-system': FileSystem,
-    'artifactory': Artifactory,
-}
-r"""Backend registry."""
-
-
-def create(
-        name: str,
-        host: str,
-        repository: str,
-) -> Backend:
-    r"""Create backend.
-
-    Args:
-        name: backend registry name
-        host: host address
-        repository: repository name
-
-    Returns:
-        backend object
-
-    Raises:
-        ValueError: if registry name does not exist
-
-    """
-    if name not in backend_registry:
-        raise ValueError(
-            'A backend class with name '
-            f"'{name} "
-            'does not exist.'
-            "Use 'register_backend()' to register one."
-        )
-
-    if name not in backends:
-        backends[name] = {}
-    if host not in backends[name]:
-        backends[name][host] = {}
-    if repository not in backends[name][host]:
-        backend = backend_registry[name]
-        backends[name][host][repository] = backend(host, repository)
-    return backends[name][host][repository]
-
-
-def register(
-        name: str,
-        cls: typing.Type[Backend],
-):
-    r"""Register backend.
-
-    If a backend with this name already exists,
-    it will be overwritten.
-
-    Args:
-        name: backend registry name
-        cls: backend class
-
-    """
-    backend_registry[name] = cls
-
-
-register('artifactory', Artifactory)
-register('file-system', FileSystem)
